using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using System.IO.Ports;

namespace TinyNetworkAnalyzer
{
    public class Measurement {
        private const double AD5933_INTOSC_FREQ = 16.776e6;

        public SerialPort serial_port;
        public Action<int> iteration_function = null;
        public double start_freq;
        public double stop_freq;
        public int point_n;
        public int settling_cycles;
        public int source_level;
        public int[] receiver_gain = new int[2];
        public enum measure_val { A, B, B_OVER_A, IMPEDANCE };
        public measure_val measure;
        public enum dumptype_val { ROW, FORMATTED, ROW_AND_FORMATTED };
        public dumptype_val dumptype;
        public enum clock_source_val { INT, EXT };
        public clock_source_val clock_source;
        public double external_clock_frequency;
        public double current_sense_r;
        public double[]  freq;      // frequency of point[point]
        public double[,] real;      // real part [channel][point]
        public double[,] imag;      // imaginary part [channel][point]
        public double[] mag;        // magnitude [point]
        public double[] arg;        // argument [point]
        public double[] z_real;     // real part of impedance [point]
        public double[] z_imag;     // imaginary part of impedance [point]
        public bool log_x;

//        public Measurement()
//        {
//        }

        public void DumpDataFileAsCSV(string filename)
        {
            string str;
            int i;

            StreamWriter writer = new StreamWriter(filename);

            str = "Frequency [Hz]";
            if((dumptype == dumptype_val.ROW) || (dumptype == dumptype_val.ROW_AND_FORMATTED)) {
                if((measure == measure_val.B_OVER_A) || (measure == measure_val.IMPEDANCE))
                    str += ", Real(Ach), Image(Ach), Real(Bch), Image(Bch)";
                else
                    str += ", Real, Image";
            }
            if((dumptype == dumptype_val.FORMATTED) || (dumptype == dumptype_val.ROW_AND_FORMATTED)) {
                if(measure == measure_val.IMPEDANCE)
                    str += ", Resistance(R)[ohm], Reactance(X)[ohm]";
                else
                    str += ", Magnitude[dB], Phase[deg]";
            }
            writer.WriteLine(str);

            for(i=0; i!=point_n; i++) {
                str = freq[i].ToString();
                if((dumptype == dumptype_val.ROW) || (dumptype == dumptype_val.ROW_AND_FORMATTED)) {
                    if((measure == measure_val.B_OVER_A) || (measure == measure_val.IMPEDANCE)) {
                        str += ", " + real[0, i].ToString() + ", " + imag[0, i].ToString();
                        str += ", " + real[1, i].ToString() + ", " + imag[1, i].ToString();
                    } else
                        str += ", " + real[0, i].ToString() + ", " + imag[0, i].ToString();
                }
                if((dumptype == dumptype_val.FORMATTED) || (dumptype == dumptype_val.ROW_AND_FORMATTED)) {
                    if(measure == measure_val.IMPEDANCE)
                        str += ", " + z_real[i].ToString() + ", " + z_imag[i].ToString();
                    else
                        str += ", " + mag[i].ToString() + ", " + arg[i].ToString();
                }
                writer.WriteLine(str);
            }
            writer.Close();
        }

        public void DumpDataFileAsGNUPLOT(string filename)
        {
            string str;
            double f1;
            int i;

            StreamWriter writer = new StreamWriter(filename);
            writer.WriteLine("# This file generated by TinyNetworkAnalyzer\n");
            if(log_x) {
                writer.WriteLine("set log xx2");
                writer.WriteLine("set grid xtics ytics x2tics");
                f1 = Math.Pow(10, (int)Math.Log10(start_freq));
                str = "set x2tics (" + start_freq.ToString() + ", ";
                while(f1 < stop_freq) {
                    if(f1 > start_freq)
                        str += ((int)f1).ToString() + ", ";
                    f1 += Math.Pow(10, (int)Math.Log10(f1));
                }
                str += stop_freq.ToString() + ")";
                writer.WriteLine(str);
            } else {
                writer.WriteLine("set grid x y");
            }
            if(measure == measure_val.IMPEDANCE) {
                writer.WriteLine("set title \"Impedance (R + jX)\"");
                writer.WriteLine("set ylabel \"Resistance / Reactance [ohms]\"");
            } else {
                writer.WriteLine("set title \"Gain & Phase\"");
                writer.WriteLine("set ylabel \"Gain [dB]\"");
                writer.WriteLine("set y2label \"Phase [deg]\"");
                writer.WriteLine("set y2tics");
            }
            writer.WriteLine("set xlabel \"Frequency [Hz]\"");
            writer.WriteLine("set format x \"%3.1s%c\"");
            if(log_x)
                writer.WriteLine("set format x2 \"\"");
            writer.WriteLine("set ytics nomirror");
            writer.WriteLine("set xrange [" + start_freq.ToString() + ":" + stop_freq.ToString() + "]");
            if(log_x)
                writer.WriteLine("set x2range [" + start_freq.ToString() + ":" + stop_freq.ToString() + "]");
            if(measure == measure_val.IMPEDANCE) {
                writer.WriteLine("plot '-' using 1:4 axis x1y1 with lines title \"Resistance(R)\", \\");
                writer.WriteLine("     '-' using 1:2 axis x1y1 with lines title \"Reactance(X)\"");
            } else {
                writer.WriteLine("plot '-' using 1:4 axis x1y1 with lines title \"Gain\", \\");
                writer.WriteLine("     '-' using 1:2 axis x1y2 with lines title \"Phase\"");
            }

            for(i=0; i!=point_n; i++) {
                str = freq[i].ToString();
                str += "\t" + real[0, i].ToString() + "\t" + imag[0, i].ToString();
                if(measure == measure_val.IMPEDANCE)
                    str += "\t" + z_real[i].ToString();
                else
                    str += "\t" + mag[i].ToString();
                writer.WriteLine(str);
            }
            writer.WriteLine("e");
            for(i=0; i!=point_n; i++) {
                if(measure == measure_val.IMPEDANCE)
                    str = freq[i].ToString() + "\t" + z_imag[i].ToString();
                else
                    str = freq[i].ToString() + "\t" + arg[i].ToString();
                writer.WriteLine(str);
            }
            writer.WriteLine("e");
            writer.Close();
        }

        public void Measure()
        {
            double freq_start, freq_inc;
            string cmd, res;
            int i;

            i = ((measure == Measurement.measure_val.B_OVER_A) || (measure == Measurement.measure_val.IMPEDANCE)) ? 2 : 1;
            freq = new double[point_n];
            real = new double[i, point_n];
            imag = new double[i, point_n];
            mag = new double[point_n];
            arg = new double[point_n];
            z_real = new double[point_n];
            z_imag = new double[point_n];

            serial_port.WriteLine("CY\r");  // Ready command
            serial_port.ReadTo("$ ");



            /*
             * setup measurement options...
             */
            cmd = "SR " + Convert.ToString(source_level) + "\r";
            serial_port.WriteLine(cmd);     // Set Range command
            serial_port.ReadTo("$ ");

            cmd = (clock_source == Measurement.clock_source_val.INT) ? "SC I\r" : "SC E\r";
            serial_port.WriteLine(cmd);     // Set Clock Source command
            serial_port.ReadTo("$ ");



            /*
             * setup frequency sweep...
             */
            double clock_freq =
                (clock_source == clock_source_val.EXT) ? external_clock_frequency : AD5933_INTOSC_FREQ;

            freq_start = start_freq / clock_freq * 0x20000000;
            freq_inc   = (stop_freq - start_freq) / (point_n - 1) /
                         clock_freq * 0x20000000;

            cmd = "FS " + Convert.ToString((int)Math.Round(freq_start), 16) + "\r";
            serial_port.WriteLine(cmd);     // Start Frequency Set command
            serial_port.ReadTo("$ ");

            cmd = "FI " + Convert.ToString((int)Math.Round(freq_inc), 16) + "\r";
            serial_port.WriteLine(cmd);     // Increment Frequency Set command
            serial_port.ReadTo("$ ");

            cmd = "FN " + Convert.ToString((int)point_n, 16) + "\r";
            serial_port.WriteLine(cmd);     // Sweep Points Set command
            serial_port.ReadTo("$ ");

            // TODO: support x2, x4 
            cmd = "SS " + Convert.ToString((int)settling_cycles, 16) + "\r";
            serial_port.WriteLine(cmd);     // Settling cycle Set command
            serial_port.ReadTo("$ ");

            serial_port.WriteLine("CC\r");  // Initialize command
            serial_port.ReadTo("$ ");



            /*
             * measurement loop...
             */
            for(i=0; i!=point_n; i++) {
                int real_i, imag_i, channel;

                freq[i] = (freq_start * clock_freq / 0x20000000) +
                          (freq_inc   * clock_freq / 0x20000000) * i;

                if(measure == measure_val.B) {
                    serial_port.WriteLine("SI B\r");    // Select input channel command
                    serial_port.ReadTo("$ ");
                    channel = 1;
                } else {
                    serial_port.WriteLine("SI A\r");    // Select input channel command
                    serial_port.ReadTo("$ ");
                    channel = 0;
                }

                cmd = (receiver_gain[channel] == 0) ? "SG 1\r" : "SG 5\r";
                serial_port.WriteLine(cmd);     // Set Gain command
                serial_port.ReadTo("$ ");

                if (i == 0) {
                    serial_port.WriteLine("CS\r");  // Start Sweep command
                    serial_port.ReadTo("$ ");
                } else {
                    serial_port.WriteLine("CI\r");  // Increment Sweep command
                    serial_port.ReadTo("$ ");
                }

                // polling for point measurement complete
                while(true) {
                    serial_port.WriteLine("RS\r");  // Read status command
                    res = serial_port.ReadTo("$ ");
                    if((Convert.ToInt32(res.Substring(5, 2), 16) & 0x02) != 0)
                        break;
                    Thread.Sleep(1);
                }

                serial_port.WriteLine("RX\r");      // Read complex data
                res = serial_port.ReadTo("$ ");
                real_i = Convert.ToInt32(res.Substring(5, 4), 16);
                imag_i = Convert.ToInt32(res.Substring(10, 4), 16);
                if(real_i > 0x8000)
                    real_i -= 0x10000;
                if(imag_i > 0x8000)
                    imag_i -= 0x10000;
                real[0, i] = (double)real_i;
                imag[0, i] = (double)imag_i;

                if((measure == measure_val.A) || (measure == measure_val.B)) {
                    // calculate Mag/Arg
                    mag[i] = 20 * Math.Log10(
                        Math.Sqrt(Math.Pow(real[0, i], 2) + Math.Pow(imag[0, i], 2)) );
                    arg[i] =
                        -Math.Atan2(imag[0, i], real[0, i]) / (2 * Math.PI) * 360;

                    // single channel only, loop complete
                    if(iteration_function != null)
                        iteration_function(i + 1);
                    continue;
                }

                serial_port.WriteLine("SI B\r");    // Select input channel command
                serial_port.ReadTo("$ ");
                channel = 1;

                cmd = (receiver_gain[channel] == 0) ? "SG 1\r" : "SG 5\r";
                serial_port.WriteLine(cmd);     // Set Gain command
                serial_port.ReadTo("$ ");

                serial_port.WriteLine("CR\r");  // Repeat Sweep command
                serial_port.ReadTo("$ ");

                // polling for point measurement complete
                while(true) {
                    serial_port.WriteLine("RS\r");  // Read status command
                    res = serial_port.ReadTo("$ ");
                    if((Convert.ToInt32(res.Substring(5, 2), 16) & 0x02) != 0)
                        break;
                    Thread.Sleep(1);
                }

                serial_port.WriteLine("RX\r");      // Read complex data
                res = serial_port.ReadTo("$ ");
                real_i = Convert.ToInt32(res.Substring(5, 4), 16);
                imag_i = Convert.ToInt32(res.Substring(10, 4), 16);
                if(real_i > 0x8000)
                    real_i -= 0x10000;
                if(imag_i > 0x8000)
                    imag_i -= 0x10000;
                real[1, i] = (double)real_i;
                imag[1, i] = (double)imag_i;

                double tmp_mag, tmp_arg;
                tmp_mag =
                    Math.Sqrt(Math.Pow(real[1, i], 2) + Math.Pow(imag[1, i], 2)) /
                    Math.Sqrt(Math.Pow(real[0, i], 2) + Math.Pow(imag[0, i], 2));
                tmp_arg =
                    -Math.Atan2(imag[1, i], real[1, i]) +
                     Math.Atan2(imag[0, i], real[0, i]) ;
                if(tmp_arg > Math.PI)
                    tmp_arg -= 2 * Math.PI;
                if(tmp_arg <= -Math.PI)
                    tmp_arg += 2 * Math.PI;
                mag[i] = 20 * Math.Log10(tmp_mag);
                arg[i] = tmp_arg / (2 * Math.PI) * 360;

                z_real[i] = ((1/tmp_mag) * Math.Cos(-tmp_arg) - 1) * current_sense_r;
                z_imag[i] =  (1/tmp_mag) * Math.Sin(-tmp_arg)      * current_sense_r;

                if(iteration_function != null)
                    iteration_function(i + 1);
            }
        }
    };
}
