using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Diagnostics; 
using System.IO;

namespace TinyNetworkAnalyzer
{
    public partial class Form1 : Form
    {
        private Form_GraphOption graph_option_dialog = new Form_GraphOption();
        private Form_SystemOption system_option_dialog = new Form_SystemOption();
        private Process gnuplot_proc = new Process();
        private string temp_gnuplot_filename = null;
        private Measurement meas = new Measurement();

        public Form1()
        {
            InitializeComponent();
            Application.ApplicationExit += new EventHandler(Application_ApplicationExit);

            string[] port_names = System.IO.Ports.SerialPort.GetPortNames();
            comboBox_ComPort.Items.AddRange(port_names);

            // remember previous used COM port from the registry
            comboBox_ComPort.SelectedIndex = 0;
            Microsoft.Win32.RegistryKey regkey =
                Microsoft.Win32.Registry.CurrentUser.OpenSubKey(@"TinyNetworkAnalyzer\test\sub", false);
            if(regkey != null) {
                string prev_com_port = (string)regkey.GetValue("COMPort");
                if(prev_com_port != null) {
                    int com_index = comboBox_ComPort.FindString(prev_com_port, -1);
                    if(com_index > 0)
                        comboBox_ComPort.SelectedIndex = com_index;
                }
                regkey.Close();
            }
        }

        private void Application_ApplicationExit(object sender, EventArgs e)
        {
            try {
                gnuplot_proc.Kill();
            } catch {
            }
            if(temp_gnuplot_filename != null)
                File.Delete(temp_gnuplot_filename);
        }

        private void button_Open_Click(object sender, EventArgs e)
        {
            serialPort1.PortName = (string)comboBox_ComPort.SelectedItem;
            try {
                serialPort1.Open();
            } catch {
                MessageBox.Show("Cannot open " + serialPort1.PortName, "Error");
                return;
            }

            // save COM port to the registry
            Microsoft.Win32.RegistryKey regkey =
                Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"TinyNetworkAnalyzer\test\sub");
            regkey.SetValue("COMPort", (string)comboBox_ComPort.SelectedItem);
            regkey.Close();
            
            // TODO: flush buffer

            comboBox_ComPort.Enabled = false;
            button_Open.Enabled = false;
            button_Close.Enabled = true;
            button_Start.Enabled = true;
        }

        private void button_Close_Click(object sender, EventArgs e)
        {
            try {
                serialPort1.Close();
            } catch {
                MessageBox.Show("Communication error.", "Error");
            }
            comboBox_ComPort.Enabled = true;
            button_Open.Enabled = true;
            button_Close.Enabled = false;
            button_Start.Enabled = false;
        }
 
        private void checkBox_GainA_CheckedChanged(object sender, EventArgs e)
        {
            if(((CheckBox)sender).Checked) {
                radioButton_GainA1.Enabled = false;
                radioButton_GainA5.Enabled = false;
                radioButton_GainA1.Checked = radioButton_GainB1.Checked;
                radioButton_GainA5.Checked = radioButton_GainB5.Checked;
            } else {
                radioButton_GainA1.Enabled = true;
                radioButton_GainA5.Enabled = true;
            }
        }

        private void radioButton_GainB_CheckedChanged(object sender, EventArgs e)
        {
            if(checkBox_GainA.Checked) {
                radioButton_GainA1.Checked = radioButton_GainB1.Checked;
                radioButton_GainA5.Checked = radioButton_GainB5.Checked;
            }
        }

        private void radioButton_Impedance_CheckedChanged(object sender, EventArgs e)
        {
            if(((RadioButton)sender).Checked)
                textBox_CurrentDetectR.Enabled = true;
            else
                textBox_CurrentDetectR.Enabled = false;
        }

        private void button_Start_Click(object sender, EventArgs e)
        {
            byte[] txbuf = new byte[1];
            double start_freq, stop_freq;
            int point_n, settling_cycles;

            try {
                if(textBox_StartFreq.Text.EndsWith("k", true, null))
                    // support "kilo" suffix
                    start_freq = Convert.ToDouble(textBox_StartFreq.Text.Substring(0, textBox_StartFreq.Text.Length-1)) * 1000;
                else
                    start_freq = Convert.ToDouble(textBox_StartFreq.Text);
                if(textBox_StopFreq.Text.EndsWith("k", true, null))
                    // support "kilo" suffix
                    stop_freq = Convert.ToDouble(textBox_StopFreq.Text.Substring(0, textBox_StopFreq.Text.Length-1)) * 1000;
                else
                    stop_freq = Convert.ToDouble(textBox_StopFreq.Text);
                point_n = Convert.ToInt32(textBox_Points.Text);
                settling_cycles = Convert.ToInt32(textBox_SettlingCycles.Text);
            } catch {
                MessageBox.Show("Invalid start/stop frequency, points, or settling cycles", "Error");
                return;
            }

            if((start_freq < 0.1) || (start_freq > 100000)) {
                MessageBox.Show("Start frequency must be in range of 0.1 to 100000", "Error");
                return;
            }

            if((stop_freq < 0.1) || (stop_freq > 100000)) {
                MessageBox.Show("Stop frequency must be in range of 0.1 to 100000", "Error");
                return;
            }

            if(stop_freq < start_freq) {
                MessageBox.Show("Stop frequency must greater than start frequency", "Error");
                return;
            }

            if((point_n < 2) || (point_n > 511)) {
                MessageBox.Show("Points must be in range of 2 to 511", "Error");
                return;
            }

            if((settling_cycles < 0) || (settling_cycles > 255)) {
                MessageBox.Show("Start frequency must be in range of 0 to 255", "Error");
                return;
            }

            meas.serial_port     = serialPort1;
            meas.iteration_function = iteration_progress;
            meas.start_freq      = start_freq;
            meas.stop_freq       = stop_freq;
            meas.point_n         = point_n;
            meas.settling_cycles = settling_cycles;
            meas.source_level = radioButton_SignalLevel0.Checked ? 0 :
                                radioButton_SignalLevel1.Checked ? 1 :
                                radioButton_SignalLevel2.Checked ? 2 : 3;
            meas.clock_source = system_option_dialog.use_external_clock ?
                Measurement.clock_source_val.EXT : Measurement.clock_source_val.INT;
            meas.external_clock_frequency = system_option_dialog.external_clock_frequency;
            meas.receiver_gain[0] = radioButton_GainA1.Checked ? 0 : 1;
            meas.receiver_gain[1] = radioButton_GainB1.Checked ? 0 : 1;
            meas.measure = radioButton_A.Checked      ? Measurement.measure_val.A :
                           radioButton_B.Checked      ? Measurement.measure_val.B :
                           radioButton_BoverA.Checked ? Measurement.measure_val.B_OVER_A :
                                                        Measurement.measure_val.IMPEDANCE;
            if(radioButton_Impedance.Checked) {
                try {
                    if(textBox_CurrentDetectR.Text.EndsWith("k", true, null))
                        meas.current_sense_r = Convert.ToDouble(
                            textBox_CurrentDetectR.Text.Substring(0, textBox_CurrentDetectR.Text.Length-1)) * 1000;
                    else
                        meas.current_sense_r = Convert.ToDouble(textBox_CurrentDetectR.Text);
                } catch {
                    MessageBox.Show("Invalid current sense R value.", "Error");
                    return;
                }
            }

            progressBar1.Visible = true;
            progressBar1.Maximum = point_n;
            progressBar1.Value = 0;
            label_Progress.Text = "0/" + point_n.ToString();
            try {
                meas.Measure();
            } catch(TimeoutException) {
                    MessageBox.Show("Communication timeout.", "Error");
                    progressBar1.Visible = false;
                    label_Progress.Text = "";
                    return;
            } catch(IOException) {
                    MessageBox.Show("Communication I/O error.", "Error");
                    progressBar1.Visible = false;
                    label_Progress.Text = "";
                    return;
            }
            progressBar1.Visible = false;
            label_Progress.Text = "";

            if(graph_option_dialog.use_gnuplot) {
                if(temp_gnuplot_filename != null)
                    File.Delete(temp_gnuplot_filename);
                temp_gnuplot_filename = Path.GetTempFileName();
                meas.log_x = graph_option_dialog.log_x;
                meas.DumpDataFileAsGNUPLOT(temp_gnuplot_filename);
                try {
                    gnuplot_proc.Kill();
                } catch {
                }
                gnuplot_proc.StartInfo.FileName = graph_option_dialog.gnuplot_exe_path;
                gnuplot_proc.StartInfo.Arguments = "-persist \"" + temp_gnuplot_filename + "\"";
                try {
                    gnuplot_proc.Start();
                } catch {
                    MessageBox.Show("Failed to call the external GNUPLOT executable.", "Error");
                }
            }
        }

        private void iteration_progress(int point)
        {
            progressBar1.PerformStep();
            label_Progress.Text = point.ToString() + "/" + progressBar1.Maximum.ToString();
            Refresh();
        }

        private void toolStripMenuItem_ExportCSV_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();

            sfd.Filter = "CSV File(*.csv)|*.csv|All Files(*.*)|*.*";
            sfd.FilterIndex = 0;
            sfd.RestoreDirectory = true;
            if (sfd.ShowDialog() == DialogResult.OK) {
                meas.dumptype = Measurement.dumptype_val.ROW_AND_FORMATTED;
                meas.DumpDataFileAsCSV(sfd.FileName);
            }
        }

        private void toolStripMenuItem_ExportGNUPLOT_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();

            sfd.Filter = "GNUPLOT Script(*.plt)|*.plt|All Files(*.*)|*.*";
            sfd.FilterIndex = 0;
            sfd.RestoreDirectory = true;
            if (sfd.ShowDialog() == DialogResult.OK) {
                meas.log_x = graph_option_dialog.log_x;
                meas.DumpDataFileAsGNUPLOT(sfd.FileName);
            }
        }

        private void toolStripMenuItem_Exit_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void toolStripMenuItem_Graph_Click(object sender, EventArgs e)
        {
            if(graph_option_dialog.ShowDialog() == DialogResult.OK) {
            }
        }

        private void toolStripMenuItem_System_Click(object sender, EventArgs e)
        {
            if(system_option_dialog.ShowDialog() == DialogResult.OK) {
            }
        }

        private void toolStripMenuItem_About1_Click(object sender, EventArgs e)
        {
            Form_About about = new Form_About();
            about.ShowDialog();
        }
    }
}