/*
 * command.c --- Command parser.
 * $Id$
 */
#pragma MESSAGE DISABLE C4000
#pragma MESSAGE DISABLE C1420
#pragma MESSAGE DISABLE C3303

#include "derivative.h"
#include "types.h"
#include "command.h"
#include "i2c.h"
#include "usb_cdc.h"

#define LINEBUF_LENGTH		16
static UInt8 s_linebuf[LINEBUF_LENGTH];
static UInt8 s_linebuf_index;
static UInt8 s_ctrl_reg[2];

static const char *s_error_message =
"Invalid command line.\r\n$ ";
static const char *s_help_message =
"  Commands\r\n"
"    FS <24bit-HEX> : Set start frequency\r\n"
"    FI <24bit-HEX> : Set frequency increment\r\n"
"    FN <16bit-HEX> : Set number of increment\r\n"
"    SS <16bit-HEX> : Set settling time cycles\r\n"
"    SR 0|1|2|3     : Set output scale (0:200mVp-p, 3:2Vp-p)\r\n"
"    SG 1|5         : Set PGA gain\r\n"
"    SC I|E         : Set clock source (I:Int, E:Ext)\r\n"
"    SI A|B         : Set input channel select\r\n"
"    CA             : Reset command\r\n"
"    CC             : Initialize command\r\n"
"    CS             : Start frequency sweep command\r\n"
"    CI             : Increment frequency command\r\n"
"    CR             : Repeat frequency command\r\n"
"    CT             : Temperature measurement command\r\n"
"    CY             : Standby command\r\n"
"    RR             : Read real data register\r\n"
"    RI             : Read image data register\r\n"
"    RX             : Read real & image data registers\r\n"
"    RT             : Read temperature data register\r\n"
"    RS             : Read status register\r\n"
"    ?              : Show this message\r\n"
"\r\n"
"$ ";

static void  put_byte(UInt8 c);
static UInt16 strlen(const UInt8 *str);
static UInt8 *skip_space(const UInt8 *buf);
static UInt8 hex_convert(const UInt8 *str, UInt8 max_digits, UInt16 *val_h, UInt16 *val_l);
static void  hex_ascii8(UInt8 val);



/*
 * Initialize
 */
void Command_Init(void)
{
	s_linebuf_index = 0;

	// set control register default value
	s_ctrl_reg[0] = 0x03;	// Vout=200mVp-p, PGA=x1
	s_ctrl_reg[1] = 0x00;	// clock source: Internal
}



/*
 *
 */
void Command_MainProcess(void)
{
	while(CDC_IsRxDataReady()) {
		char c;
		c = CDC_GetByte();

		if(c == 0x08) {
			if(s_linebuf_index != 0) {
				// Backspace
				s_linebuf_index--;
				put_byte(0x08);
				put_byte(' ');
				put_byte(0x08);
			}
			return;
		}

		if(c == 0x0d) {
			// CR
			s_linebuf[s_linebuf_index] = '\0';
			s_linebuf_index = 0;
			put_byte(0x0d);
			put_byte(0x0a);
			Command_ParseCommand(s_linebuf);
			return;
		}

		if(s_linebuf_index < (LINEBUF_LENGTH - 1)) {
			s_linebuf[s_linebuf_index++] = c;
			put_byte(c);
		}
	}
}



/*
 * Command-line parser
 */
void Command_ParseCommand(UInt8 *line)
{
	UInt8 *l;
	UInt8 i;
	UInt16 val_h, val_l;
	UInt8 ctrl_reg[2];

	// to upper
	l = line;
	while(*l != 0) {
		if((*l >= 0x61) && (*l <= 0x7a))
			*l -= 0x20;
		l++;
	}

	// skip LF code
	l = line;
	while(*l == 0x0a)
		l++;

	switch(*l++) {
		case 'F':
			// Frequency command
			if(*l == 'S') {
				// Start frequency command
				l = skip_space(l+1);
				i = hex_convert(l, 6, &val_h, &val_l);
				if(i == 0)
					goto error;
				I2C_ByteWrite(0x82, val_h & 0xff);
				I2C_ByteWrite(0x83, val_l >> 8);
				I2C_ByteWrite(0x84, val_l & 0xff);
			} else if(*l == 'I') {
				// Increment frequency command
				l = skip_space(l+1);
				i = hex_convert(l, 6, &val_h, &val_l);
				if(i == 0)
					goto error;
				I2C_ByteWrite(0x85, val_h & 0xff);
				I2C_ByteWrite(0x86, val_l >> 8);
				I2C_ByteWrite(0x87, val_l & 0xff);
			} else if(*l == 'N') {
				// Number of frequency increment command
				l = skip_space(l+1);
				i = hex_convert(l, 4, &val_h, &val_l);
				if(i == 0)
					goto error;
				I2C_ByteWrite(0x88, val_l >> 8);
				I2C_ByteWrite(0x89, val_l & 0xff);
			} else
				// error
				goto error;

			put_byte('$');
			put_byte(' ');

			break;

		case 'S':
			// Setup command
			if(*l == 'S') {
				// Settling time command
				l = skip_space(l+1);
				i = hex_convert(l, 4, &val_h, &val_l);
				if(i == 0)
					goto error;
				I2C_ByteWrite(0x8a, val_l >> 8);
				I2C_ByteWrite(0x8b, val_l & 0xff);
			} else if(*l == 'R') {
				// Range of output command
				l = skip_space(l+1);
				if(*l == '0')
					s_ctrl_reg[0] = (s_ctrl_reg[0] & 0xf9) | 0x02;	// 200mVp-p
				else if(*l == '1')
					s_ctrl_reg[0] = (s_ctrl_reg[0] & 0xf9) | 0x04;	// 400mVp-p
				else if(*l == '2')
					s_ctrl_reg[0] = (s_ctrl_reg[0] & 0xf9) | 0x06;	// 1Vp-p
				else if(*l == '3')
					s_ctrl_reg[0] = s_ctrl_reg[0] & 0xf9;	// 2Vp-p
				else
					goto error;
				I2C_ByteWrite(0x80, s_ctrl_reg[0]);
			} else if(*l == 'G') {
				// Gain of PGA command
				l = skip_space(l+1);
				if(*l == '1')
					s_ctrl_reg[0] |= 0x01;
				else if(*l == '5')
					s_ctrl_reg[0] &= 0xfe;
				else
					goto error;
				I2C_ByteWrite(0x80, s_ctrl_reg[0]);
			} else if(*l == 'C') {
				// Clock source select
				l = skip_space(l+1);
				if(*l == 'I')
					s_ctrl_reg[1] &= 0xf7;
				else if(*l == 'E')
					s_ctrl_reg[1] |= 0x08;
				else
					goto error;
				I2C_ByteWrite(0x81, s_ctrl_reg[1]);
			} else if(*l == 'I') {
				// Input channel select
				l = skip_space(l+1);
				if(*l == 'A') {
					PTAD = (PTAD & 0xf7) & 0xfc;	// PTA[3]=Low,  PTA[1:0]=Low for I2C
					PTAD = (PTAD | 0x04) & 0xfc;	// PTA[2]=High, PTA[1:0]=Low for I2C
				} else if(*l == 'B') {
					PTAD = (PTAD & 0xfb) & 0xfc;	// PTA[2]=Low,  PTA[1:0]=Low for I2C
					PTAD = (PTAD | 0x08) & 0xfc;	// PTA[3]=High, PTA[1:0]=Low for I2C
				} else
					goto error;
			} else
				// error
				goto error;

			put_byte('$');
			put_byte(' ');

			break;

		case 'C':
			// Control command
			ctrl_reg[0] = s_ctrl_reg[0];
			ctrl_reg[1] = s_ctrl_reg[1];
			if(*l == 'A') {
				// All reset command
				s_ctrl_reg[0] = 0x03;	// initial value: Vout=200mVp-p, PGA=x1
				s_ctrl_reg[1] = 0x00;	// initial value: clock source: Internal
				I2C_ByteWrite(0x81, 0x10);	// Reset
			} else if(*l == 'C')
				// Clear command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0x10);
			else if(*l == 'S')
				// Start command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0x20);
			else if(*l == 'I')
				// Increment command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0x30);
			else if(*l == 'R')
				// Repeat command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0x40);
			else if(*l == 'T')
				// Temperature measurement command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0x90);
			else if(*l == 'Y')
				// Standby command
				I2C_ByteWrite(0x80, s_ctrl_reg[0] | 0xb0);
			else
				// error
				goto error;

			put_byte('$');
			put_byte(' ');

			break;

		case 'R':
			// Read command
			if(*l == 'R') {
				// Real data register read command
				hex_ascii8(I2C_ByteRead(0x94));
				hex_ascii8(I2C_ByteRead(0x95));
			} else if(*l == 'I') {
				// Image data register read command
				hex_ascii8(I2C_ByteRead(0x96));
				hex_ascii8(I2C_ByteRead(0x97));
			} else if(*l == 'X') {
				// Real & image data registers read command
				hex_ascii8(I2C_ByteRead(0x94));
				hex_ascii8(I2C_ByteRead(0x95));
				put_byte(' ');
				hex_ascii8(I2C_ByteRead(0x96));
				hex_ascii8(I2C_ByteRead(0x97));
			} else if(*l == 'T') {
				// Temprature data register read command
				hex_ascii8(I2C_ByteRead(0x92));
				hex_ascii8(I2C_ByteRead(0x93));
			} else if(*l == 'S') {
				// Status register read command
				hex_ascii8(I2C_ByteRead(0x8f));
			} else
				// error
				goto error;

			put_byte('\r');
			put_byte('\n');
			put_byte('$');
			put_byte(' ');

			break;

		case '?':
			// Help command
			l = (UInt8 *)s_help_message;
			while(*l != 0)
				put_byte(*l++);
			break;

		case '\0':
		case ' ':
			// empty
			put_byte('$');
			put_byte(' ');
			break;

		default:
			// error
			goto error;
	}

	return;

error:
	l = (UInt8 *)s_error_message;
	while(*l != 0)
		put_byte(*l++);
}



/*
 * Put byte with TX buffer check & iteration
 */
static void put_byte(UInt8 c)
{
	while(!CDC_IsTxDataReady())
		CDC_MainProcess();
	CDC_PutByte(c);
}



/*
 * Return string length
 */
static UInt16 strlen(const UInt8 *str)
{
	UInt16 len = 0;

	while(*str++ != '\0')
		len++;

	return len;
}



/*
 * Skip space
 */
static UInt8 *skip_space(const UInt8 *buf)
{
	while(*buf == ' ')
		buf++;

	return (UInt8 *)buf;
}



/*
 * Decode hex value
 */
static UInt8 hex_convert(const UInt8 *str, UInt8 max_digits, UInt16 *val_h, UInt16 *val_l)
{
	UInt8 c;
	UInt8 cnv_digits = 0;
	UInt8 tmp;
	UInt16 tmp_val_h = 0, tmp_val_l = 0;

	while(1) {
		c = *str++;
		if((c == ' ') || (c == '\0')) {
			*val_h = tmp_val_h;
			*val_l = tmp_val_l;

			return cnv_digits;
		}

		if(max_digits-- == 0)
			return 0;	// over digits

		// shift one digit
		tmp = tmp_val_l >> 12;
		tmp_val_l <<= 4;
		tmp_val_h <<= 4;
		tmp_val_h |= tmp;

		if((c >= '0') && (c <= '9'))
			tmp_val_l |= c - '0';
		else if((c >= 'A') && (c <= 'F'))
			tmp_val_l |= c - 'A' + 10;
		else
			return 0;	// error

		cnv_digits++;
	}
}



/*
 * 1byte hex to ascii
 */
static void hex_ascii8(UInt8 val)
{
	UInt8 c;

	if((val >> 4) >= 10)
		c = (val >> 4) - 10 + 'A';
	else
		c = (val >> 4) + '0';
	put_byte(c);

	val &= 0x0f;
	if(val >= 10)
		c = val - 10 + 'A';
	else
		c = val + '0';
	put_byte(c);
}