/*
 * usb.h --- USB function foundation.
 * $Id$
 */

#include "derivative.h"
#include "types.h"
#include "irq.h"
#include "usb.h"
#include "usb_config.h"

/*
 * EP control information struct
 */
typedef struct {
	UInt8 bufsize;
	UInt16 tlength;
	UInt16 maxlength;
	UInt8 state;
	UInt8 *address;
	UInt8 flags;
	UInt8 data0;
} EP_INFO_T;

#define EP_STATE_IDLE			0x01
#define EP_STATE_DATA_TX		0x02
#define EP_STATE_DATA_TX_LAST	0x04
#define EP_STATE_DATA_RX		0x08
#define EP_STATE_STATUS_TX		0x10
#define EP_STATE_STATUS_RX		0x20
#define EP_STATE_TX_STOP		0x40

#define EP_FLAG_ZPACKET			0x01

/*
 * indexed EPCTL register
 */
#define EPCTL					(&EPCTL0)

/*
 * static variables
 */
static UInt8 s_usb_current_config;
static UInt8 s_usb_state;
static UInt8 s_new_address;
static EP_INFO_T s_ep_info[TOTAL_NUMBER_OF_EP];
static USB_EVENT_HANDLER_SET_T s_usb_event_handlers;

/*
 * EP number to BDT index convert table
 */
#define EP_INDEX_IN				0
#define EP_INDEX_OUT			1

static const UInt8 s_ep_to_bdt_index_table[5][2] = {
	{ 0, 1 },	/* EP0 IN,OUT */
	{ 2, 2 },	/* EP1 */
	{ 3, 3 },	/* EP2 */
	{ 4, 4 },	/* EP3 */
	{ 5, 5 },	/* EP4 */
};

/*
 * EP EPADR value table
 */
static const UInt8 s_epadr_table[] = {
	USB_GET_EPADR(USB_EP1_BUF_BASE),	/* EP1 */
	USB_GET_EPADR(USB_CDC_TX_BUF_BASE),	/* EP2 */
	USB_GET_EPADR(USB_CDC_RX_BUF_BASE)	/* EP3 */
};

/*
 * internal functions
 */
static void  usb_send_sub(UInt8 ep);
static void  usb_receive_sub(UInt8 ep);
static void  usb_ep0_setup_handler(void);
static UInt8 usb_setup_ep(UInt8 addr, UInt8 type, UInt8 ep, UInt8 bufsize);
static void  usb_reset_handler_internal(void);
static void  ep0_handler(void);



void USB_Init(USB_EVENT_HANDLER_SET_T *usb_event_handlers)
{
	// copy handler pointer
	s_usb_event_handlers.usb_reset_handler  = usb_event_handlers->usb_reset_handler;
	s_usb_event_handlers.usb_sleep_handler  = usb_event_handlers->usb_sleep_handler;
	s_usb_event_handlers.usb_resume_handler = usb_event_handlers->usb_resume_handler;
	s_usb_event_handlers.usb_ep0_handler    = usb_event_handlers->usb_ep0_handler;

	// reset USB module
	USBCTL0_USBRESET = 1;
	while(USBCTL0_USBRESET != 0);

	// setup the ep0
	s_ep_info[0].state = EP_STATE_IDLE;
	s_ep_info[0].flags = 0;
	s_ep_info[0].tlength = 0;
	s_ep_info[0].maxlength = 0;
	s_ep_info[0].bufsize = EP0_BUFFER_SIZE;

	BDT_ADDR(0) = USB_GET_EPADR(USB_EP0_TX_BUF_BASE);
	BDT_BYTECOUNT(0) = EP0_BUFFER_SIZE;
	BDT_CTRL(0) = 0;

	BDT_ADDR(1) = USB_GET_EPADR(USB_EP0_RX_BUF_BASE);
	BDT_BYTECOUNT(1) = EP0_BUFFER_SIZE;
	BDT_CTRL(1) = BDT_CTRL_OWN | BDT_CTRL_DTS;

	EPCTL[0] = EPCTL0_EPRXEN_MASK | EPCTL0_EPTXEN_MASK | EPCTL0_EPHSHK_MASK;

	// enable USB PHY
	USBCTL0 = USBCTL0_USBPHYEN_MASK | USBCTL0_USBVREN_MASK;

	// clear interrupt status except USBRSTF flag
	INTSTAT = ~INTSTAT_USBRSTF_MASK;

	// disable error interrupt
	ERRENB = 0x00;

	// enable usb reset interrupt
	INTENB_USBRST = 1;

	// enable USB
	CTL_USBEN = 1;

	// indicate attach status
	USBCTL0_USBPU = 1;
}



void USB_Send(UInt8 ep, const UInt8 *data, UInt16 tr_length, UInt16 req_length)
{
	UInt8 imask;

	s_ep_info[ep].tlength = (tr_length < req_length) ? tr_length : req_length;
	s_ep_info[ep].maxlength = s_ep_info[ep].tlength;
	s_ep_info[ep].address = (UInt8 *)data;
	s_ep_info[ep].flags = (tr_length < req_length) ? EP_FLAG_ZPACKET : 0;
	s_ep_info[ep].state = EP_STATE_DATA_TX;

	imask = irq_save_and_disable();

	usb_send_sub(ep);

	irq_restore(imask);
}



void USB_Receive(UInt8 ep, UInt8 *data, UInt16 tr_length)
{
	UInt8 bdt_index;

	s_ep_info[ep].tlength = tr_length;
	s_ep_info[ep].maxlength = s_ep_info[ep].tlength;
	s_ep_info[ep].address = data;
	s_ep_info[ep].state = EP_STATE_DATA_RX;
	s_ep_info[ep].flags = 0;

	// setup the EP for receive
	bdt_index = s_ep_to_bdt_index_table[ep][EP_INDEX_OUT];
	BDT_BYTECOUNT(bdt_index) = s_ep_info[ep].bufsize;
	BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[ep].data0 | BDT_CTRL_DTS;

	s_ep_info[ep].data0 ^= BDT_CTRL_DATA01;
}



UInt8 USB_GetState(void)
{
	return s_usb_state;
}



UInt16 USB_GetDone(UInt8 ep)
{
	return s_ep_info[ep].maxlength - s_ep_info[ep].tlength;
}



UInt8 USB_EPIsBusy(UInt8 ep)
{
	return (s_ep_info[ep].state != EP_STATE_IDLE) ? 1 : 0;
}



static UInt8 usb_setup_ep(UInt8 addr, UInt8 type, UInt8 ep, UInt8 bufsize)
{
	// disable endpoint
	EPCTL[ep] = 0x00;

	s_ep_info[ep].state = EP_STATE_IDLE;
	s_ep_info[ep].flags = 0;
	s_ep_info[ep].tlength = 0;
	s_ep_info[ep].maxlength = 0;
	s_ep_info[ep].bufsize = bufsize;

	if(type == EP_TYPE_DISABLE)
		return 0;

	if(ep == 0) {
		// configure EP0 TX(IN) BDT
		BDT_ADDR(0) = USB_GET_EPADR(USB_EP0_TX_BUF_BASE);
		BDT_BYTECOUNT(0) = bufsize;
		BDT_CTRL(0) = 0;

		// configure EP0 RX(OUT) BDT
		BDT_ADDR(1) = USB_GET_EPADR(USB_EP0_RX_BUF_BASE);
		BDT_BYTECOUNT(1) = bufsize;
		BDT_CTRL(1) = BDT_CTRL_OWN | BDT_CTRL_DTS;

		EPCTL[0] = EPCTL0_EPRXEN_MASK | EPCTL0_EPTXEN_MASK | EPCTL0_EPHSHK_MASK;

		// clear suspend state
		CTL_TSUSPEND = 0;
	} else {
		BDT_ADDR(ep + 1) = s_epadr_table[ep - 1];
		BDT_BYTECOUNT(ep + 1) = bufsize;
		BDT_CTRL(ep + 1) = 0x00;

		if(addr & 0x80)
			EPCTL[ep] = EPCTL0_EPTXEN_MASK | EPCTL0_EPHSHK_MASK;
		else
			EPCTL[ep] = EPCTL0_EPRXEN_MASK | EPCTL0_EPHSHK_MASK;
	}

	return 0;
}



static void set_config(UInt8 cfg_ndx)
{
	UInt8 iface, ep;

	CTL_ODDRST=1;

	s_usb_current_config = cfg_ndx;

	CTL_ODDRST=0;

	if((cfg_ndx != 0) && (cfg_ndx <= MAX_CONFIG)) {
		for(iface=0; iface!=MAX_INTERFACE(cfg_ndx); iface++) {
			for(ep=0; ep!=MAX_EP(cfg_ndx, iface); ep++) {
				const UInt8 *epd = SELECT_EP_DESCRIPTOR(iface, ep);

				(void)usb_setup_ep(epd[2], epd[3], (UInt8)(epd[2] & 0x7fu), epd[4]);
			}
		}

		s_usb_state = USB_DEVICE_STATE_CONFIGURED;
	} else
		s_usb_state = USB_DEVICE_STATE_ADDRESSED;
}



static void usb_reset_handler_internal(void)
{
	s_usb_current_config = 0;
	s_usb_state = USB_DEVICE_STATE_ADDRESSED;

	// reset to default address
	ADDR = 0;

	// disable all endpoints except EP0
	EPCTL[1] = 0x00;
	EPCTL[2] = 0x00;
	EPCTL[3] = 0x00;
	EPCTL[4] = 0x00;
	EPCTL[5] = 0x00;
	EPCTL[6] = 0x00;

	// flush any pending transactions
	while(INTSTAT_TOKDNEF)
		INTSTAT = 0xbf;

	s_usb_state = USB_DEVICE_STATE_DEFAULT;

	// enable USB interrupts
	INTENB = INTENB_USBRST_MASK | INTENB_TOKDNE_MASK | INTENB_SLEEP_MASK;
}



static void usb_send_sub(UInt8 ep)
{
	UInt8 bdt_index;
	UInt8 len;
	UInt8 *p;
	UInt8 i;

	bdt_index = s_ep_to_bdt_index_table[ep][EP_INDEX_IN];

	switch (s_ep_info[ep].state) {
	case EP_STATE_DATA_TX:
		len = (s_ep_info[ep].tlength < s_ep_info[ep].bufsize) ?
			(UInt8)s_ep_info[ep].tlength : s_ep_info[ep].bufsize;

		// copy data from user buffer
		p = (UInt8 *)( BDT_BASE + (BDT_ADDR(bdt_index) << 2) );
		for(i=0; i!=len; i++)
			*p++ = *s_ep_info[ep].address++;

		// prepare for transmit
		BDT_BYTECOUNT(bdt_index) = len;
		BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[ep].data0 | BDT_CTRL_DTS;

		s_ep_info[ep].data0 ^= BDT_CTRL_DATA01;	// invert data toggle bit
		s_ep_info[ep].tlength -= len;

		// last packet ?
		if(s_ep_info[ep].tlength == 0) {
			if((len == s_ep_info[ep].bufsize) && (s_ep_info[ep].flags & EP_FLAG_ZPACKET))
				s_ep_info[ep].state = EP_STATE_TX_STOP;	// next state, create zero size packet
			else
				s_ep_info[ep].state = EP_STATE_DATA_TX_LAST;	// this is last packet
		}

		break;

	case EP_STATE_TX_STOP:
		// create zero size packet
		s_ep_info[ep].state = EP_STATE_DATA_TX_LAST;	// this is last packet
		BDT_BYTECOUNT(bdt_index) = 0;
		BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[ep].data0 | BDT_CTRL_DTS;
		break;

	case EP_STATE_DATA_TX_LAST:
		// complete the last packet transmit
		s_ep_info[ep].state = EP_STATE_IDLE;
		break;

	default:
		break;
	}
}



void usb_receive_sub(UInt8 ep)
{
	UInt8 bdt_index;
	UInt8 len;
	UInt8 *p;
	UInt8 i;

	bdt_index = s_ep_to_bdt_index_table[ep][EP_INDEX_OUT];
	len = BDT_BYTECOUNT(bdt_index);	// actual received length

	// copy data to user buffer
	p = (UInt8 *)( BDT_BASE + (BDT_ADDR(bdt_index) << 2) );
	for(i=0; i!=len; i++)
		*(UInt8 *)s_ep_info[ep].address++ = *p++;

	s_ep_info[ep].tlength -= len;

	if((s_ep_info[ep].tlength == 0) || (s_ep_info[ep].bufsize != len))
		// complete data packet
		s_ep_info[ep].state = EP_STATE_IDLE;
}



interrupt 7 static void usb_it_handler(void)
{
	UInt8 iflags;

	iflags = INTSTAT & INTENB;

	if(iflags & INTSTAT_USBRSTF_MASK) {
		INTSTAT_USBRSTF = 1;	// clear interrupt flag

		// enter default state
		usb_reset_handler_internal();
		if(s_usb_event_handlers.usb_reset_handler != NULL)
			s_usb_event_handlers.usb_reset_handler();
	}

	if(iflags & INTSTAT_SLEEPF_MASK) {
		INTENB_SLEEP = 0;
		INTSTAT_SLEEPF = 1;	// clear interrupt flag
		INTENB_RESUME = 1;
		if(s_usb_event_handlers.usb_sleep_handler != NULL)
			s_usb_event_handlers.usb_sleep_handler();
	}

	if(iflags & INTSTAT_RESUMEF_MASK) {
		INTENB_RESUME = 0;
		INTSTAT_RESUMEF = 1;	// clear interrupt flag
		INTENB_SLEEP = 1;
		if(s_usb_event_handlers.usb_resume_handler != NULL)
			s_usb_event_handlers.usb_resume_handler();
	}

	if(iflags & INTSTAT_TOKDNEF_MASK) {
		UInt8 ep = STAT_ENDP;

		// EP0 packet ?
		if(ep == 0)
			ep0_handler();
		else {
			if(STAT_IN)
				// IN (TX)
				usb_send_sub(ep);
			else
				// OUT (RX)
				usb_receive_sub(ep);
		}

		INTSTAT_TOKDNEF = 1;	// clear interrupt flag
	}
}



static void ep0_handler()
{
	UInt8 *pdata = (UInt8 *)USB_EP0_RX_BUF_BASE;
	UInt8  reqtype = pdata[USB_SETUP_INDEX_REQTYPE];
	UInt8  req     = pdata[USB_SETUP_INDEX_REQ];
	UInt16 value   = ((UInt16)pdata[USB_SETUP_INDEX_VALUE_H] << 8) |
	                 pdata[USB_SETUP_INDEX_VALUE_L];
	UInt16 index   = ((UInt16)pdata[USB_SETUP_INDEX_INDEX_H] << 8) |
	                 pdata[USB_SETUP_INDEX_INDEX_L];
	UInt16 length  = ((UInt16)pdata[USB_SETUP_INDEX_LENGTH_H] << 8) |
	                 pdata[USB_SETUP_INDEX_LENGTH_L];
	UInt8 bdt_index;

	if( (s_ep_info[0].state & (EP_STATE_DATA_TX | EP_STATE_DATA_TX_LAST | EP_STATE_TX_STOP)) &&
	    !STAT_IN)
		// we are sending the data, however host send the data
		// abort transmit and start receive the data
		s_ep_info[0].state = EP_STATE_IDLE;

	switch(s_ep_info[0].state) {
	case EP_STATE_IDLE:
		if((BDT_CTRL(1) & BDT_CTRL_BDTKPID) == (SETUP_TOKEN << 2)) {
			// SETUP token received
			CTL_TSUSPEND = 0;	/* XXX this point?? */

			// after the setup packet, its next packet is always DATA1
			s_ep_info[0].data0 = BDT_CTRL_DATA01;

			usb_ep0_setup_handler();
			if(reqtype & 0x80) {
				// TX(IN): prepare buffer for receive
				bdt_index = s_ep_to_bdt_index_table[0][EP_INDEX_OUT];
				BDT_BYTECOUNT(bdt_index) = s_ep_info[0].bufsize;
				BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[0].data0 | BDT_CTRL_DTS;
			} else {
				// RX(OUT)
				if(s_ep_info[0].state == EP_STATE_IDLE)
					goto send_status;
			}
		}

		break;

	case EP_STATE_DATA_TX:
	case EP_STATE_DATA_TX_LAST:
	case EP_STATE_TX_STOP:
		if(s_ep_info[0].state == EP_STATE_DATA_TX_LAST)
			break;

		usb_send_sub(0);

		// if transfer is ended, we shall receive a handshake from the host
		if(s_ep_info[0].state == EP_STATE_IDLE)
			s_ep_info[0].state = EP_STATE_STATUS_TX;

		break;

	case EP_STATE_STATUS_TX:
		// when IN transfer, we received a zero byte DATA1 packet
		s_ep_info[0].state = EP_STATE_IDLE;

		// prepare buffer for receive
		bdt_index = s_ep_to_bdt_index_table[0][EP_INDEX_OUT];
		BDT_BYTECOUNT(bdt_index) = s_ep_info[0].bufsize;
		BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[0].data0 | BDT_CTRL_DTS;

		break;

	case EP_STATE_DATA_RX:
		usb_receive_sub(0);

send_status:
		if(s_ep_info[0].state == EP_STATE_IDLE) {
			// last packet received
			UInt8 bdt_index;

			// handshake is allways DATA1
			s_ep_info[0].data0 = BDT_CTRL_DATA01;

			// current status is sending state of OUT transaction
			s_ep_info[0].state = EP_STATE_STATUS_RX;

			// send a zero byte data1 packet as response
			bdt_index = s_ep_to_bdt_index_table[0][EP_INDEX_IN];
			BDT_BYTECOUNT(bdt_index) = 0;
			BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[0].data0 | BDT_CTRL_DTS;
		}

		break;

	case EP_STATE_STATUS_RX:
		// when OUT transfer, we sent a zero byte DATA0 packet
		s_ep_info[0].state = EP_STATE_IDLE;
		
		// if device state is ADDR_PENDING, update device address and state
		if(s_usb_state == USB_DEVICE_STATE_ADDR_PENDING) {
			if(s_new_address != 0) {
				s_usb_state = USB_DEVICE_STATE_ADDRESSED;
				ADDR = s_new_address;
			} else
				s_usb_state = USB_DEVICE_STATE_DEFAULT;
		}

		// prepare buffer for receive
		bdt_index = s_ep_to_bdt_index_table[0][EP_INDEX_OUT];
		BDT_BYTECOUNT(bdt_index) = s_ep_info[0].bufsize;
		BDT_CTRL(bdt_index) = BDT_CTRL_OWN | s_ep_info[0].data0 | BDT_CTRL_DTS;

		break;

	default:
		break;
	}
}



static void usb_ep0_setup_handler(void)
{
	UInt8 *pdata = (UInt8 *)USB_EP0_RX_BUF_BASE;
	UInt8  reqtype = pdata[USB_SETUP_INDEX_REQTYPE];
	UInt8  req     = pdata[USB_SETUP_INDEX_REQ];
	UInt16 value   = ((UInt16)pdata[USB_SETUP_INDEX_VALUE_H] << 8) |
	                 pdata[USB_SETUP_INDEX_VALUE_L];
	UInt16 index   = ((UInt16)pdata[USB_SETUP_INDEX_INDEX_H] << 8) |
	                 pdata[USB_SETUP_INDEX_INDEX_L];
	UInt16 length  = ((UInt16)pdata[USB_SETUP_INDEX_LENGTH_H] << 8) |
	                 pdata[USB_SETUP_INDEX_LENGTH_L];

	if((reqtype & 0x7f) != 0x00) {
		// not the standard device request, shirk to the user EP0 handler
		if(s_usb_event_handlers.usb_ep0_handler != NULL)
			s_usb_event_handlers.usb_ep0_handler();
	} else {
		switch(req) {
		case USBRQ_SET_ADDRESS:
			s_new_address = (UInt8)(value & 0x7f);
			s_usb_state = USB_DEVICE_STATE_ADDR_PENDING;
			break;

		case USBRQ_GET_DESCRIPTOR:
			switch(value >> 8) {
			case STDD_DEVICE:
				// get device descriptor
				USB_Send(0, (UInt8 *)usb_device_descriptor,
					usb_device_descriptor[0], length);
				break;

			case STDD_CONFIG:
				// get configuration descriptor
				USB_Send(0, (UInt8 *)usb_config_descriptor,
					((UInt16)usb_config_descriptor[3] << 8) | usb_config_descriptor[2],
					length);
				break;

			case STDD_STRING:
				// get string descriptor
				USB_Send(0, (UInt8 *)usb_string_descriptors[value & 0xff],
					usb_string_descriptors[value & 0xff][0], length);
				break;

			default:
				break;
			}
			break;

		case USBRQ_GET_CONFIGURATION:
			USB_Send(0, (void *)&s_usb_current_config, 1, length);
			break;

		case USBRQ_SET_CONFIGURATION:
			set_config(value & 0xff);
			break;

		default:
			break;
		}
	}
}